/* Message indexing
 *
 * Author: Sam Magnuson <zachsman@wiw.org>
 * This code is under GPL
 *
 */
#ifndef kmmsgindex_h
#define kmmsgindex_h

#include <qobject.h>
#include <qmap.h>
#include <qstring.h>
#include <qintdict.h>
#include <qdatetime.h>
#include <qvaluelist.h>
#include <qvaluestack.h>
#include <qguardedptr.h>

#include "kmfolder.h"

class KMMsgIndexRef;
class KMIndexSearchTarget;
class KMSearchPattern;
class KMSearchRule;
class KMSearch;

class KMMsgIndex : public QObject
{
    Q_OBJECT

public:
    static const int kmindex_grow_increment;
    typedef enum { INDEX_IDLE = 1, INDEX_RESTORE, INDEX_CREATE } IndexState;
    typedef enum
    {
        HEADER_BYTEORDER = 0,
	HEADER_VERSION = 1,
	HEADER_COMPLETE = 2,
	HEADER_COUNT = 3,
	HEADER_USED = 4,
	HEADER_INDEXED = 5,
	HEADER_REMOVED = 6,
	HEADER_end = 7
    } MsgHeader;

    typedef enum
    {
      CHUNK_HEADER_COUNT = 0,
      CHUNK_HEADER_USED = 1,
      CHUNK_HEADER_NEXT = 2,
      CHUNK_HEADER_end = 3
    } ChunkHeader;

    typedef enum
    {
      TOC_BODY = 0,
      TOC_HEADER_NAME = 1,
      TOC_HEADER_DATA = 2
    } TOC;

    struct TermIndex
    {
        int fd;
	Q_INT32 count;
	Q_INT32 used;
	Q_INT32 indexed;
	Q_INT32 removed;
	KMMsgIndexRef* ref;
	QString loc;
    };
    struct H
    {
        int next_hnum;
	QMap<QCString, Q_UINT16> header_lookup;
	QMap<Q_UINT16, QMap<QCString, int> > headers;
    };
    struct TermTOC
    {
        int fd;
	QString loc;
	QMap<QCString, int> body;
	H h;
    };
    struct TermProcessed
    {
        int fd;
	QString loc;
	QIntDict<void> known;
    };
    struct Create
    {
        int timer_id;
	QValueStack<Q_UINT32> serNums;
	QValueStack<QGuardedPtr<KMFolder> > folders;
    };
    struct Restore
    {
        int timer_id;
	bool reading_processed;
	bool restart_index;
    };

private:
    QValueList<QGuardedPtr<KMFolder> > mOpenedFolders;

    //book-keeping
    TermIndex mTermIndex;
    TermTOC mTermTOC;
    TermProcessed mTermProcessed;
    uchar_t mIndexState;
    // IndexState mIndexState;

    //state information
    Create create;
    Restore restore;
    int delay_cnt;
    QTime mLastSearch; //when last index lookup was performed
    QIntDict<KMIndexSearchTarget> mActiveSearches; //for async search

public:
    KMMsgIndex(QObject *parent=NULL, const char *name=NULL);
    virtual ~KMMsgIndex() { reset(); }
    
    void init();
    void remove();

    bool canHandleQuery(KMSearchRule *);
    bool canHandleQuery(KMSearchPattern *);
    QValueList<Q_UINT32> query(KMSearchRule *, bool full_phrase_search = true);
    QValueList<Q_UINT32> query(KMSearchPattern *, bool full_phras_search = true);

    //convenient..
    bool startQuery(KMSearch *);
    bool stopQuery(KMSearch *);
    bool stopQuery(int id) { return mActiveSearches.remove(id); }

    //force the cleanup..
    void cleanUp();

private:
    //internal
    bool recreateIndex();
    void readIndex();
    void syncIndex();

    void flush();
    void reset(bool = true);

    bool createState(bool = true);
    bool restoreState(bool = true);

    int addBucket(int, Q_UINT32);
    QValueList<Q_UINT32> find(QString, bool, KMSearchRule *, bool = false,
			      Q_UINT16 = 0);
    void values(int, int, QValueList<Q_UINT32> &);
    void values(int, int, QIntDict<void> &);

    int processMsg(Q_UINT32);
    bool addHeaderTerm(Q_UINT16, const char *, uchar, Q_UINT32);
    bool addBodyTerm(const char *, uchar, Q_UINT32);
    int allocTermChunk(int);

    bool isTimeForClean();
    bool isKillTerm(const char *, uchar);
    bool isKillHeader(const char *, uchar);

protected:
    void timerEvent(QTimerEvent *);

private slots:
    void slotAddMsg(KMFolder *, Q_UINT32);
    void slotRemoveMsg(KMFolder *, Q_UINT32);
};

class KMIndexSearchTarget : public QObject
{
    Q_OBJECT
    int mId;
    bool mVerifyResult;
    QValueList<QGuardedPtr<KMFolder> > mOpenedFolders;
    QValueStack<Q_UINT32> mSearchResult;
    QGuardedPtr<KMSearch> mSearch;

public:
    KMIndexSearchTarget(KMSearch *);
    virtual ~KMIndexSearchTarget();

    void stop(bool stop_search=true)
    {
	if (stop_search && mSearch)
	    emit proxyFinished(true);
	mSearch = NULL;
    }
    KMSearch *search() const { return mSearch; }
    int id() const { return mId; }

protected:
    void timerEvent(QTimerEvent *);

signals:
    void proxyFound(Q_UINT32);
    void proxyFinished(bool);
};

#endif /* __KMMSGINDEX_H__ */










