#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <strings.h>
#include <udpmessagesocket.h>


void UDPMessageSocket::reInit() {
	if (socketfd!=-1) close(socketfd);
	type = SocketUDP;
	if ( ( socketfd = socket( AF_INET6, SOCK_DGRAM, 0 ) ) == -1 ) {
		if ( ( socketfd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 ) {
			perror( "UDPMessageSocket::UDPMessageSocket(): socket() failed" );
			intype=InetNone;
		} else { 
			intype=Inet4;
		}
	} else {
		intype=Inet6;
	}
	didcomplain = false;
	printf ("UDPMessageSocket::UDPMessageSocket(): socketfd=%d (%s)\n",socketfd,intype==Inet6?"IPv6":intype==Inet4?"IPv4":"None");
}


UDPMessageSocket::UDPMessageSocket( void )
{
	reInit();
}


void UDPMessageSocket::reInit( int newfd )
{
	struct sockaddr soa,*soap=&soa;
	size_t l=sizeof(soa);

	if (socketfd!=-1) close(socketfd);
	if ( getsockname(newfd, soap, &l) == -1 ) {
		perror( "UDPMessageSocket::UDPMessageSocket(): getsockname() failed" );
		intype = Inet4;
	} else {
		switch ( ((struct sockaddr *)soap)->sa_family ) {
		case AF_INET  : intype = Inet4; break;
		case AF_INET6 : intype = Inet6; break;
		default       : intype = InetNone; 
		}
	}
	
	type = SocketUDP;
	socketfd = newfd;
	didcomplain = false;
	printf ("UDPMessageSocket::UDPMessageSocket(%d): socketfd=%d (%s)\n",newfd,socketfd,intype==Inet6?"IPv6":intype==Inet4?"IPv4":"None");
}


UDPMessageSocket::UDPMessageSocket( int newfd )
{
	reInit(newfd);
}


UDPMessageSocket::~UDPMessageSocket( void )
{
	close( socketfd );
}


int UDPMessageSocket::connect( unsigned int portnum )
{
	if (intype==Inet6) {
		bzero( &socketaddress.soa6, sizeof( sockaddr_in6 ) ); 
		socketaddress.soa6.sin6_port = htons( portnum );
		socketaddress.soa6.sin6_family = AF_INET6;
		bcopy ( &ip_addr.in6_addr, &(socketaddress.soa6.sin6_addr), 16);
		socketaddress.soa6.sin6_flowinfo = 0;
		socketaddress.soa6.sin6_scope_id = 0;
	} else {
		socketaddress.soa4.sin_port = htons( portnum );
		socketaddress.soa4.sin_family = AF_INET;
		socketaddress.soa4.sin_addr = ip_addr.in4_addr;
		bzero( &(socketaddress.soa4.sin_zero), 8 ); // is this portable?
	}

	remoteaddress = socketaddress;
	return 0;
}


int UDPMessageSocket::SetTOS( void )
{
	unsigned char tos;
	socklen_t optlen;
	tos=IPTOS_PREC_CRITIC_ECP;
	optlen=1;
	if(setsockopt(socketfd,SOL_IP,IP_TOS,&tos,optlen) != 0){
		perror("UDPMessageSocket::SetTOS()");
	}
	return 0;
}


void hexdump(unsigned char *p, int len, int wordsize) {
  long n;
  int c;
  char t[17];

  if (!p) { printf ("hexdump: NULL pointer\n");   return; }
  for (n=0, c=*p++; len>0 || n%16!=0; len--) {
    if (!(n&15)) printf ("%08X:",(unsigned int)n);
    if (!(n%wordsize)) printf (" ");
    printf ((len>0)?"%02X":"  ",c);
    t[n%16]=c<32?'.':c>126?'-':c;    t[n%16+1]=0;
    if (n++%16==15) printf ("  %s\n",t);
    c=(len>1)?*p++:32;
  }

}


int UDPMessageSocket::send( const char *sendbuffer, unsigned int length )
{
	int numbytes;
	int soal=intype==Inet6?sizeof(struct sockaddr_in6):intype==Inet4?sizeof(struct sockaddr_in):0;

//	printf ("UDPMessageSocket::send() : intype=%s\n",intype==Inet6?"Inet6":intype==Inet4?"Inet4":"InetNone");
//	printf ("UDPMessageSocket::send() : socketfd=%d\n",socketfd);
//	printf ("UDPMessageSocket::send() : sendbuffer=%p\n",sendbuffer);
//	printf ("UDPMessageSocket::send() : length=%d\n",length);
//	printf ("UDPMessageSocket::send() : hexdump of sockaddr:\n");
//	hexdump((unsigned char *)(&remoteaddress),sizeof(struct sockaddr_in6),1);
	if ( ( numbytes = sendto( socketfd, sendbuffer, length, 0, &remoteaddress.soa, soal ) ) == -1 ) {
		if( !didcomplain ) {
			perror( "UDPMessageSocket::send(): sendto() failed" );
			didcomplain = true;
		}
		return -1;
	}
	return 0;
}


int UDPMessageSocket::receive( char *recvbuffer, unsigned int maxlength )
{
	socklen_t addrlength;
	int numbytes;
	addrlength = sizeof( struct sockaddr );
	if ( ( numbytes = recvfrom( socketfd, recvbuffer, maxlength, 0, &remoteaddress.soa, &addrlength ) ) == -1 ) {
		perror( "UDPMessageSocket::receive(): recvfrom() failed" );
		return -1;
	}
	return numbytes;
}


unsigned int UDPMessageSocket::listen( unsigned int portnum )
{
	struct sockaddr name;
	socklen_t namesize;
	int count = 0;
	u_int16_t *pn;
	size_t soal=0;

	if (intype==Inet6) {
		soal = sizeof( struct sockaddr_in6 );
		bzero( &socketaddress.soa6, soal ); 
		socketaddress.soa6.sin6_port = htons( portnum );
		socketaddress.soa6.sin6_family = AF_INET6;
		pn = &socketaddress.soa6.sin6_port;
		printf ("Binding socket to [::]:%d\n",portnum);
	} else {
		socketaddress.soa4.sin_port = htons( portnum );
		socketaddress.soa4.sin_family = AF_INET;
		socketaddress.soa4.sin_addr.s_addr = INADDR_ANY;
		soal = sizeof( struct sockaddr_in );
		pn = &socketaddress.soa4.sin_port;
		bzero( &(socketaddress.soa4.sin_zero), 8 ); // is this portable?
		printf ("Binding socket to 0.0.0.0:%d\n",portnum);
	}

	while ( bind( socketfd, &socketaddress.soa, soal ) == -1  && count <= 10 ) {
		count++;
		portnum += 2;
		*pn = htons( portnum );
	}
	if ( count > 10 ) {
		perror( "UDPMessageSocket::listen(): bind() failed" );
		printf ("#: EINVAL=%d   EBADF=%d   ENOTSOCK=%d\n",EINVAL,EBADF,ENOTSOCK);
		printf ("#: errno==%d   socketfd=%d\n",errno,socketfd);
		printf ("#: AF_INET=%d   AF_INET6=%d\n",AF_INET,AF_INET6);
		printf ("#: sa_family: %d   intype:%d\n",socketaddress.soa.sa_family,intype);
		printf ("#: sin6_port=%d\n", ntohs(socketaddress.soa6.sin6_port));
		printf ("#: soal=%d\n",soal);
		hexdump((unsigned char *)&socketaddress.soa6, sizeof (struct sockaddr_in6),1);
		return 0;
	}
	namesize = sizeof( struct sockaddr );
	bzero( &name, sizeof( struct sockaddr ) );
	if ( getsockname( socketfd, &name, &namesize ) == -1 ) {
		perror( "UDPMessageSocket::listen(): getsockname() failed" );
		return 0;
	}
	
	ourport = ntohs( ((struct sockaddr_in *) &name)->sin_port );
	printf ("#: ourport=%d\n",ourport);
	return ourport;
}


int UDPMessageSocket::listenOnEvenPort( void )
{
	struct sockaddr name;
	socklen_t namesize;
	int oldsocketfd = 0;
	int portnum = 0;
	size_t soal;
	
getaddress:
	if (intype==Inet6) {
		soal = sizeof( sockaddr_in6 );
		bzero( &socketaddress.soa6, soal ); 
		socketaddress.soa6.sin6_port = htons( portnum );
		socketaddress.soa6.sin6_family = AF_INET6;
	} else {
		socketaddress.soa4.sin_port = htons( portnum );
		socketaddress.soa4.sin_family = AF_INET;
		socketaddress.soa4.sin_addr.s_addr = INADDR_ANY;
		soal = sizeof( sockaddr_in );
		bzero( &(socketaddress.soa4.sin_zero), 8 ); // is this portable?
	}

	if ( bind( socketfd, &socketaddress.soa, soal ) == -1 ) {
		perror( "UDPMessageSocket::listen(): bind() failed" );
		return -1;
	}
	namesize = sizeof( struct sockaddr );
	bzero( &name, sizeof( struct sockaddr ) );
	if ( getsockname( socketfd, &name, &namesize ) == -1 ) {
		perror( "UDPMessageSocket::listen(): getsockname() failed" );
		return -1;
	}
	ourport = ntohs( ((struct sockaddr_in *) &name)->sin_port );
	printf( "UDPMessageSocket: Listening on %d\n", ourport );
	if( ( portnum == 0 ) && ( ourport % 2 ) != 0 ) {
		printf( "UDPMessageSocket: Retrying...\n" );
		oldsocketfd = socketfd;
		if (intype==InetNone) return -1;
		if (intype==Inet4) {
			if( ( socketfd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 ) {
				perror( "UDPMessageSocket::UDPMessageSocket(): socket() failed" );
			}
		}
		if (intype==Inet6) {
			if( ( socketfd = socket( AF_INET6, SOCK_DGRAM, 0 ) ) == -1 ) {
				perror( "UDPMessageSocket::UDPMessageSocket(): socket() failed" );
			}
		}
		goto getaddress;
	}
	if( oldsocketfd != 0 ) close( oldsocketfd );
	return 0;
}


int UDPMessageSocket::listenOnEvenPortForVideo( void )
{
	struct sockaddr name;
	socklen_t namesize;
	int oldsocketfd = 0;
	int portnum = 0;

	if (intype==InetNone) return -1;
	if (intype==Inet4) {
		if( ( socketfd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 ) {
			perror( "UDPMessageSocket::UDPMessageSocket(): socket() failed" );
		}
	}
	if (intype==Inet6) {
		if( ( socketfd = socket( AF_INET6, SOCK_DGRAM, 0 ) ) == -1 ) {
			perror( "UDPMessageSocket::UDPMessageSocket(): socket() failed" );
		}
	}

	size_t soal;
	
getaddress:
	if (intype==Inet6) {
		soal = sizeof( sockaddr_in6 );
		bzero( &socketaddress.soa6, soal ); 
		socketaddress.soa6.sin6_port = htons( portnum );
		socketaddress.soa6.sin6_family = AF_INET6;
	} else {
		socketaddress.soa4.sin_port = htons( portnum );
		socketaddress.soa4.sin_family = AF_INET;
		socketaddress.soa4.sin_addr.s_addr = INADDR_ANY;
		soal = sizeof( sockaddr_in );
		bzero( &(socketaddress.soa4.sin_zero), 8 ); // is this portable?
	}

	int one = 1;
	if ( setsockopt( socketfd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof( one ) ) == -1 ) {
		perror( "UDPMessageSocket::setsockopt SO_REUSEADDR" );
	}

	if ( bind( socketfd, &socketaddress.soa, soal ) == -1 ) {
		perror( "UDPMessageSocket::listen(): bind() failed" );
		return -1;
	}

	namesize = sizeof( struct sockaddr );
	bzero( &name, sizeof( struct sockaddr ) );
	if ( getsockname( socketfd, &name, &namesize ) == -1 ) {
		perror( "UDPMessageSocket::listen(): getsockname() failed" );
		return -1;
	}
	ourport = ntohs( ((struct sockaddr_in *) &name)->sin_port );
	printf( "UDPMessageSocket (video): Listening on %d\n", ourport );
	if( portnum == 0 && ( ourport % 4 ) != 0 ) {
		printf( "UDPMessageSocket: Retrying...\n" );
		oldsocketfd = socketfd;
		if (intype==InetNone) return -1;
		if (intype==Inet4) {
			if( ( socketfd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 ) {
				perror( "UDPMessageSocket::UDPMessageSocket(): socket() failed" );
			}
		}
		if (intype==Inet6) {
			if( ( socketfd = socket( AF_INET6, SOCK_DGRAM, 0 ) ) == -1 ) {
				perror( "UDPMessageSocket::UDPMessageSocket(): socket() failed" );
			}
		}
		goto getaddress;
	}
	if( oldsocketfd != 0 ) close( oldsocketfd );

	return 0;
}


int UDPMessageSocket::accept( void )
{
	return socketfd;
}

